package xyz.lychee.lagfixer.libs.adventure.platform.bukkit;

import com.viaversion.viaversion.api.Via;
import com.viaversion.viaversion.api.connection.UserConnection;
import java.lang.invoke.MethodHandle;
import java.util.Collection;
import java.util.Iterator;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.SoundCategory;
import org.bukkit.boss.BarColor;
import org.bukkit.boss.BarFlag;
import org.bukkit.boss.BarStyle;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.entity.Player;
import org.bukkit.util.Vector;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Color;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Overlay;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.key.Key;
import xyz.lychee.lagfixer.libs.adventure.permission.PermissionChecker;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.Facet;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.FacetBase;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.FacetPointers;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.Knob;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.FacetPointers.Type;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointer;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointers;
import xyz.lychee.lagfixer.libs.adventure.sound.SoundStop;
import xyz.lychee.lagfixer.libs.adventure.sound.Sound.Source;
import xyz.lychee.lagfixer.libs.adventure.text.Component;
import xyz.lychee.lagfixer.libs.adventure.util.TriState;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;

class BukkitFacet<V extends CommandSender> extends FacetBase<V> {
   protected BukkitFacet(final @Nullable Class<? extends V> viewerClass) {
      super(viewerClass);
   }

   static final class PlayerPointers extends BukkitFacet<Player> implements Facet.Pointers<Player> {
      PlayerPointers() {
         super(Player.class);
      }

      public void contributePointers(final Player viewer, final Pointers.Builder builder) {
         Pointer var10001 = Identity.UUID;
         Objects.requireNonNull(viewer);
         builder.withDynamic(var10001, viewer::getUniqueId);
         builder.withDynamic(Identity.DISPLAY_NAME, () -> {
            return BukkitComponentSerializer.legacy().deserializeOrNull(viewer.getDisplayName());
         });
         builder.withStatic(FacetPointers.TYPE, Type.PLAYER);
         builder.withDynamic(FacetPointers.WORLD, () -> {
            return Key.key(viewer.getWorld().getName());
         });
      }
   }

   static final class ConsoleCommandSenderPointers extends BukkitFacet<ConsoleCommandSender> implements Facet.Pointers<ConsoleCommandSender> {
      ConsoleCommandSenderPointers() {
         super(ConsoleCommandSender.class);
      }

      public void contributePointers(final ConsoleCommandSender viewer, final Pointers.Builder builder) {
         builder.withStatic(FacetPointers.TYPE, Type.CONSOLE);
      }
   }

   static final class CommandSenderPointers extends BukkitFacet<CommandSender> implements Facet.Pointers<CommandSender> {
      CommandSenderPointers() {
         super(CommandSender.class);
      }

      public void contributePointers(final CommandSender viewer, final Pointers.Builder builder) {
         Pointer var10001 = Identity.NAME;
         Objects.requireNonNull(viewer);
         builder.withDynamic(var10001, viewer::getName);
         builder.withStatic(PermissionChecker.POINTER, (perm) -> {
            if (viewer.isPermissionSet(perm)) {
               return viewer.hasPermission(perm) ? TriState.TRUE : TriState.FALSE;
            } else {
               return TriState.NOT_SET;
            }
         });
      }
   }

   static final class TabList extends Message<Player> implements Facet.TabList<Player, String> {
      private static final boolean SUPPORTED = MinecraftReflection.hasMethod(Player.class, "setPlayerListHeader", String.class);

      TabList() {
         super(Player.class);
      }

      public boolean isSupported() {
         return SUPPORTED && super.isSupported();
      }

      public void send(final Player viewer, final @Nullable String header, final @Nullable String footer) {
         if (header != null && footer != null) {
            viewer.setPlayerListHeaderFooter(header, footer);
         } else if (header != null) {
            viewer.setPlayerListHeader(header);
         } else if (footer != null) {
            viewer.setPlayerListFooter(footer);
         }

      }
   }

   static final class ViaHook implements Function<Player, UserConnection> {
      public UserConnection apply(final @NotNull Player player) {
         return Via.getManager().getConnectionManager().getConnectedClient(player.getUniqueId());
      }
   }

   static class BossBar extends Message<Player> implements Facet.BossBar<Player> {
      protected final org.bukkit.boss.BossBar bar;

      protected BossBar(final @NotNull Collection<Player> viewers) {
         super(Player.class);
         this.bar = Bukkit.createBossBar("", BarColor.PINK, BarStyle.SOLID, new BarFlag[0]);
         this.bar.setVisible(false);
         Iterator var2 = viewers.iterator();

         while(var2.hasNext()) {
            Player viewer = (Player)var2.next();
            this.bar.addPlayer(viewer);
         }

      }

      public void bossBarInitialized(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar) {
         super.bossBarInitialized(bar);
         this.bar.setVisible(true);
      }

      public void bossBarNameChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final @NotNull Component oldName, final @NotNull Component newName) {
         if (!this.bar.getPlayers().isEmpty()) {
            this.bar.setTitle(this.createMessage((Player)this.bar.getPlayers().get(0), newName));
         }

      }

      public void bossBarProgressChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final float oldPercent, final float newPercent) {
         this.bar.setProgress((double)newPercent);
      }

      public void bossBarColorChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color oldColor, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color newColor) {
         BarColor color = this.color(newColor);
         if (color != null) {
            this.bar.setColor(color);
         }

      }

      private @Nullable BarColor color(final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color color) {
         if (color == Color.PINK) {
            return BarColor.PINK;
         } else if (color == Color.BLUE) {
            return BarColor.BLUE;
         } else if (color == Color.RED) {
            return BarColor.RED;
         } else if (color == Color.GREEN) {
            return BarColor.GREEN;
         } else if (color == Color.YELLOW) {
            return BarColor.YELLOW;
         } else if (color == Color.PURPLE) {
            return BarColor.PURPLE;
         } else if (color == Color.WHITE) {
            return BarColor.WHITE;
         } else {
            Knob.logUnsupported(this, color);
            return null;
         }
      }

      public void bossBarOverlayChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay oldOverlay, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay newOverlay) {
         BarStyle style = this.style(newOverlay);
         if (style != null) {
            this.bar.setStyle(style);
         }

      }

      private @Nullable BarStyle style(final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay overlay) {
         if (overlay == Overlay.PROGRESS) {
            return BarStyle.SOLID;
         } else if (overlay == Overlay.NOTCHED_6) {
            return BarStyle.SEGMENTED_6;
         } else if (overlay == Overlay.NOTCHED_10) {
            return BarStyle.SEGMENTED_10;
         } else if (overlay == Overlay.NOTCHED_12) {
            return BarStyle.SEGMENTED_12;
         } else if (overlay == Overlay.NOTCHED_20) {
            return BarStyle.SEGMENTED_20;
         } else {
            Knob.logUnsupported(this, overlay);
            return null;
         }
      }

      public void bossBarFlagsChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final @NotNull Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsAdded, final @NotNull Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsRemoved) {
         Iterator var4 = flagsRemoved.iterator();

         xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag addFlag;
         BarFlag flag;
         while(var4.hasNext()) {
            addFlag = (xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag)var4.next();
            flag = this.flag(addFlag);
            if (flag != null) {
               this.bar.removeFlag(flag);
            }
         }

         var4 = flagsAdded.iterator();

         while(var4.hasNext()) {
            addFlag = (xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag)var4.next();
            flag = this.flag(addFlag);
            if (flag != null) {
               this.bar.addFlag(flag);
            }
         }

      }

      private @Nullable BarFlag flag(final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Flag flag) {
         if (flag == Flag.DARKEN_SCREEN) {
            return BarFlag.DARKEN_SKY;
         } else if (flag == Flag.PLAY_BOSS_MUSIC) {
            return BarFlag.PLAY_BOSS_MUSIC;
         } else if (flag == Flag.CREATE_WORLD_FOG) {
            return BarFlag.CREATE_FOG;
         } else {
            Knob.logUnsupported(this, flag);
            return null;
         }
      }

      public void addViewer(final @NotNull Player viewer) {
         this.bar.addPlayer(viewer);
      }

      public void removeViewer(final @NotNull Player viewer) {
         this.bar.removePlayer(viewer);
      }

      public boolean isEmpty() {
         return !this.bar.isVisible() || this.bar.getPlayers().isEmpty();
      }

      public void close() {
         this.bar.removeAll();
      }
   }

   static class BossBarBuilder extends BukkitFacet<Player> implements Facet.BossBar.Builder<Player, BossBar> {
      private static final boolean SUPPORTED = MinecraftReflection.hasClass("org.bukkit.boss.BossBar");

      protected BossBarBuilder() {
         super(Player.class);
      }

      public boolean isSupported() {
         return super.isSupported() && SUPPORTED;
      }

      public @NotNull BossBar createBossBar(final @NotNull Collection<Player> viewers) {
         return new BossBar(viewers);
      }
   }

   static class SoundWithCategory extends Sound {
      private static final boolean SUPPORTED = MinecraftReflection.hasMethod(Player.class, "stopSound", String.class, MinecraftReflection.findClass("org.bukkit.SoundCategory"));

      public boolean isSupported() {
         return super.isSupported() && SUPPORTED;
      }

      public void playSound(final @NotNull Player viewer, final xyz.lychee.lagfixer.libs.adventure.sound.@NotNull Sound sound, final @NotNull Vector vector) {
         SoundCategory category = this.category(sound.source());
         if (category == null) {
            super.playSound(viewer, sound, vector);
         } else {
            String name = name(sound.name());
            viewer.playSound(vector.toLocation(viewer.getWorld()), name, category, sound.volume(), sound.pitch());
         }

      }

      public void stopSound(final @NotNull Player viewer, final @NotNull SoundStop stop) {
         SoundCategory category = this.category(stop.source());
         if (category == null) {
            super.stopSound(viewer, stop);
         } else {
            String name = name(stop.sound());
            viewer.stopSound(name, category);
         }

      }

      private @Nullable SoundCategory category(final xyz.lychee.lagfixer.libs.adventure.sound.Sound.@Nullable Source source) {
         if (source == null) {
            return null;
         } else if (source == Source.MASTER) {
            return SoundCategory.MASTER;
         } else if (source == Source.MUSIC) {
            return SoundCategory.MUSIC;
         } else if (source == Source.RECORD) {
            return SoundCategory.RECORDS;
         } else if (source == Source.WEATHER) {
            return SoundCategory.WEATHER;
         } else if (source == Source.BLOCK) {
            return SoundCategory.BLOCKS;
         } else if (source == Source.HOSTILE) {
            return SoundCategory.HOSTILE;
         } else if (source == Source.NEUTRAL) {
            return SoundCategory.NEUTRAL;
         } else if (source == Source.PLAYER) {
            return SoundCategory.PLAYERS;
         } else if (source == Source.AMBIENT) {
            return SoundCategory.AMBIENT;
         } else if (source == Source.VOICE) {
            return SoundCategory.VOICE;
         } else {
            Knob.logUnsupported(this, source);
            return null;
         }
      }
   }

   static class Sound extends Position implements Facet.Sound<Player, Vector> {
      private static final boolean KEY_SUPPORTED = MinecraftReflection.hasClass("org.bukkit.NamespacedKey");
      private static final boolean STOP_SUPPORTED = MinecraftReflection.hasMethod(Player.class, "stopSound", String.class);
      private static final MethodHandle STOP_ALL_SUPPORTED;

      public void playSound(final @NotNull Player viewer, final xyz.lychee.lagfixer.libs.adventure.sound.@NotNull Sound sound, final @NotNull Vector vector) {
         String name = name(sound.name());
         Location location = vector.toLocation(viewer.getWorld());
         viewer.playSound(location, name, sound.volume(), sound.pitch());
      }

      public void stopSound(final @NotNull Player viewer, final @NotNull SoundStop stop) {
         if (STOP_SUPPORTED) {
            String name = name(stop.sound());
            if (name.isEmpty() && STOP_ALL_SUPPORTED != null) {
               try {
                  STOP_ALL_SUPPORTED.invoke(viewer);
               } catch (Throwable var5) {
                  Throwable error = var5;
                  Knob.logError(error, "Could not invoke stopAllSounds on %s", new Object[]{viewer});
               }

               return;
            }

            viewer.stopSound(name);
         }

      }

      protected static @NotNull String name(final @Nullable Key name) {
         if (name == null) {
            return "";
         } else {
            return KEY_SUPPORTED ? name.asString() : name.value();
         }
      }

      static {
         STOP_ALL_SUPPORTED = MinecraftReflection.findMethod(Player.class, "stopAllSounds", Void.TYPE);
      }
   }

   static class Position extends BukkitFacet<Player> implements Facet.Position<Player, Vector> {
      protected Position() {
         super(Player.class);
      }

      public @NotNull Vector createPosition(final @NotNull Player viewer) {
         return viewer.getLocation().toVector();
      }

      public @NotNull Vector createPosition(final double x, final double y, final double z) {
         return new Vector(x, y, z);
      }
   }

   static class Chat extends Message<CommandSender> implements Facet.Chat<CommandSender, String> {
      protected Chat() {
         super(CommandSender.class);
      }

      public void sendMessage(final @NotNull CommandSender viewer, final @NotNull Identity source, final @NotNull String message, final @NotNull Object type) {
         viewer.sendMessage(message);
      }
   }

   static class Message<V extends CommandSender> extends BukkitFacet<V> implements Facet.Message<V, String> {
      protected Message(final @Nullable Class<? extends V> viewerClass) {
         super(viewerClass);
      }

      public @NotNull String createMessage(final @NotNull V viewer, final @NotNull Component message) {
         return BukkitComponentSerializer.legacy().serialize(message);
      }
   }
}
